{-# LANGUAGE RecordWildCards #-}
{-# OPTIONS_GHC -Wall #-}
module Debian.Packaging
       ( ServerPackage(..)
       , SiteSpec(..)
       , sitePackageName
       , serverPackageName
       , serviceName
       , databaseDirectory
       , apacheSiteName
       , apacheLogDirectory
       , serverLogDirectory
       , apacheErrorLog
       , apacheAccessLog
       , serverAppLog
       , serverAccessLog
       , prebuild
       , dropPrefix
       , dropSuffix
       ) where

import Data.List (isPrefixOf, isSuffixOf, nub)
import Data.Maybe (mapMaybe, fromMaybe)
import System.FilePath ((</>))
import System.Process (readProcess, showCommandForUser)
import Debian.Install (File, Flag(D), install)

-- | prebuild gets a list of the binary package names in the control
-- file and looks at the ones that have the siteName as a prefix and
-- any of the four valid suffixes.  Based on this it creates files in
-- the debian subdirectory to manage the web site.  The prebuild
-- function is normally called from the function UserHooks.postConf,
-- so that it gets executed before dpkg-buildpackage runs, like so:
-- 
-- > import Distribution.Simple
-- > import Distribution.Simple.LocalBuildInfo (LocalBuildInfo(buildDir))
-- > import Debian.Packaging (SiteSpec(..), databaseDirectory)
-- >
-- > siteSpec = SiteSpec { siteName = "mysite"
-- >                     , package = Dummy
-- >                     , domain = "mysite.com"
-- >                     , testHost = "mytesthost.com"
-- >                     , proxyPort = "9023"
-- >                     , serverAdmin = "logic@seereason.com"
-- >                     , headerMessage = "Generated by mysite/Setup.hs"
-- >                     , retry = "60" }
-- >
-- > main = defaultMainWithHooks simpleUserHooks
-- >        { postConf =
-- >             \ _ _ _ lbi ->
-- >                 -- This buildDir means this was run by dpkg-buildpackage
-- >                 case buildDir lbi of
-- >                   "dist-ghc/build" -> prebuild (\ spec -> ["--top", databaseDirectory spec, "+RTS", "-IO", "-RTS"]) siteSpec
-- >                   _ -> return () }
-- 
-- There are two modes of operation depending on an examination of the
-- package names in debian/control.
prebuild :: (SiteSpec -> [String]) -> SiteSpec -> IO ()
prebuild serverFlags siteSpec =
    readProcess "dh_listpackages" [] "" >>= mapM_ doSite . testMulti siteSpec . lines
    where
      doSite spec =
          mapM_ (uncurry (install [D]))
                (filter (not . null . snd)
                        (map (\ f -> f spec)
                             [ debianPostinst
                             , debianLinks
                             , debianFiles
                             , debianDirs
                             , debianInit (serverFlags spec)
                             , debianPostinst
                             , apacheSite
                             , logrotateConfig]))

data ServerPackage
    = Single      -- ^ A single package containing server binary and all supporting files
    | Server      -- ^ A package containing only the server executable, to be accompanied by configuration packages
    | Production  -- ^ The binary is in another package, this contains the configuration for a production server
    | Staging     -- ^ Contains the configuration for an almost-production server
    | Development -- ^ Contains the configuration for a testing server
    | Dummy       -- ^ An invalid value used to build a SiteSpec template.
    deriving (Eq, Show)

-- | Does this package contain the server executable?
needsServerExecutable :: SiteSpec -> Bool
needsServerExecutable (SiteSpec {package = p}) = elem p [Single, Server]
-- | Does this package require an apache site file?
needsApacheSite :: SiteSpec -> Bool
needsApacheSite (SiteSpec {package = p}) = elem p [Single, Production]
-- | Does this package cause an init script to start?
needsInit :: SiteSpec -> Bool
needsInit (SiteSpec {package = p}) = elem p [Single, Production, Staging, Development]

interestingPackageSuffixes :: [String]
interestingPackageSuffixes = ["-server", "-production", "-development", "-staging"]

-- | Information about the web site we are packaging.
data SiteSpec
  = SiteSpec
    { siteName :: String      -- ^ This is the base of the server's
                              -- debian binary package names, and is
                              -- used in various spots.
    , package :: ServerPackage -- ^ What sort of package this
                               -- represents - production,
                               -- development, etc.
    , domain :: String        -- ^ The domain name assigned to the
                              -- server
    , ports :: [(ServerPackage, String)]
                              -- ^ This must assign each and every
                              -- server package to a different port.
    , testHost :: String      -- ^ Web address used for test servers
    , serverAdmin :: String   -- ^ Apache ServerAdmin parameter
    , headerMessage :: String -- ^ A comment that will be inserted to
                              -- explain how the file was generated
    , retry :: String         -- ^ start-stop-daemon --retry argument
    }

executable :: SiteSpec -> String
executable site = siteName site ++ "-server"

siteSuffix :: SiteSpec -> String
siteSuffix (SiteSpec {package = Single}) = "-server"
siteSuffix (SiteSpec {package = Server}) = "-server"
siteSuffix (SiteSpec {package = Production}) = "-production"
siteSuffix (SiteSpec {package = Staging}) = "-staging"
siteSuffix (SiteSpec {package = Development}) = "-development"
siteSuffix (SiteSpec {package = Dummy}) = error "siteSuffix"

sitePackageName :: SiteSpec -> String
sitePackageName spec = siteName spec ++ siteSuffix spec
serviceName :: SiteSpec -> String
serviceName spec = sitePackageName spec -- dh_installinit forces the service name to equal the binary package name
databaseDirectory :: SiteSpec -> String
databaseDirectory spec = "/srv" </> sitePackageName spec
apacheSiteName :: SiteSpec -> String
apacheSiteName spec = domain spec
apacheLogDirectory :: SiteSpec -> String
apacheLogDirectory spec = "/var/log/apache2/" ++ sitePackageName spec
serverLogDirectory :: SiteSpec -> String
serverLogDirectory spec = "/var/log/" ++ sitePackageName spec
apacheErrorLog :: SiteSpec -> String
apacheErrorLog spec = apacheLogDirectory spec </> "error.log"
apacheAccessLog :: SiteSpec -> String
apacheAccessLog spec = apacheLogDirectory spec </> "access.log"
serverAppLog :: SiteSpec -> String
serverAppLog spec = serverLogDirectory spec </> "app.log"
serverAccessLog :: SiteSpec -> String
serverAccessLog spec = serverLogDirectory spec </> "access.log"

debianPostinst :: SiteSpec -> File
debianPostinst spec@(SiteSpec {..}) =
  ("debian" </> sitePackageName spec ++ ".postinst",
   if needsInit spec
   then ([ "#!/bin/sh"
          , ""
          , "case \"$1\" in"
          , "  configure)" ] ++
          apache ++
          [ "    service " ++ serviceName spec ++ " start"
          , "    ;;"
          , "esac" ])
   else [])
    where
      apache =
          if needsApacheSite spec
          then [ "    # Apache won't start if this directory doesn't exist"
               , "    mkdir -p " ++ apacheLogDirectory spec
               , "    # Restart apache so it sees the new file in /etc/apache2/sites-enabled"
               , "    /usr/sbin/a2enmod proxy"
               , "    /usr/sbin/a2enmod proxy_http"
               , "    service apache2 restart" ]
          else []


debianDirs :: SiteSpec -> File
debianDirs spec@(SiteSpec {..}) =
  ("debian" </> sitePackageName spec ++ ".dirs",
   if needsApacheSite spec
   then [apacheLogDirectory spec,
         "/etc/apache2/sites-available"]
   else [])

debianFiles :: SiteSpec -> File
debianFiles spec@(SiteSpec {..}) =
    ("debian" </> sitePackageName spec ++ ".install",
     logrotate ++ apacheSiteInstall ++ serverBinaryInstall)
    where
      logrotate =
          if needsInit spec
          then ["dist-ghc/logrotate" </> sitePackageName spec ++ " etc/logrotate.d"]
          else []
      apacheSiteInstall =
          if needsApacheSite spec
          then ["dist-ghc/apachesite" </> apacheSiteName spec ++ " etc/apache2/sites-available"]
          else []
      serverBinaryInstall =
          if needsServerExecutable spec
          then ["dist-ghc" </> "build" </> executable spec </> executable spec ++ " " ++ " usr/bin"]
          else []

debianLinks :: SiteSpec-> File
debianLinks spec@(SiteSpec {..}) =
    ("debian" </> sitePackageName spec ++ ".links", apacheSiteLink)
    where
      apacheSiteLink =
          if needsApacheSite spec
          then ["/etc/apache2/sites-available/" ++ apacheSiteName spec ++ " /etc/apache2/sites-enabled/" ++ apacheSiteName spec]
          else []

debianInit :: [String] -- ^ These are application specific flags to pass to the server
           -> SiteSpec -- ^ Specifications of the debian binary package we are generating
           -> File
debianInit extraServerFlags spec@(SiteSpec{..}) =
  ("debian" </> serviceName spec ++ ".init",
   if needsInit spec
   then   [ "#! /bin/sh -e"
          , ""
          , ". /lib/lsb/init-functions"
          , ""
          , "case \"$1\" in"
          , "  start)"
          , "    test -x /usr/bin/" ++ executable spec ++ " || exit 0"
          , "    log_begin_msg \"Starting " ++ sitePackageName spec ++ "...\""
          , "    mkdir -p " ++ databaseDirectory spec
          , "    " ++ startCommand
          , "    log_end_msg $?"
          , "    ;;"
          , "  stop)"
          , "    log_begin_msg \"Stopping " ++ sitePackageName spec ++ "...\""
          , "    " ++ stopCommand
          , "    log_end_msg $?"
          , "    ;;"
          , "  *)"
          , "    log_success_msg \"Usage: ${0} {start|stop}\""
          , "    exit 1"
          , "esac"
          , ""
          , "exit 0" ]
   else [])
  where
    startCommand = showCommandForUser "start-stop-daemon" (startOptions ++ commonOptions ++ ["--"] ++ serverOptions)
    stopCommand = showCommandForUser "start-stop-daemon" (stopOptions ++ commonOptions)
    commonOptions = ["--pidfile", "/var/run/" ++ sitePackageName spec]
    startOptions = ["--start", "-b", "--quiet", "--make-pidfile", "-d", databaseDirectory spec, "--exec", "/usr/bin" </> executable spec]
    stopOptions = ["--stop", "--quiet", "--retry", retry, "--oknodo"]
    serverOptions = baseURI ++ ["--http-port", port] ++ extraServerFlags
    baseURI = if needsApacheSite spec
              then [ "--base-uri", "http://" ++ domain ++ "/"]
              else if needsInit spec
                   then ["--base-uri", "http://" ++ testHost ++ ":" ++ port ++ "/"]
                   else []
    port = fromMaybe (error $ "No port assignment for server type " ++ show package) (lookup package ports)

-- | An apache site configuration file.  This is installed via a line
-- in debianFiles.
apacheSite :: SiteSpec -> File
apacheSite spec@(SiteSpec {..}) =
  ("dist-ghc/apachesite" </> apacheSiteName spec,
   if needsApacheSite spec
   then   [ "# " ++ headerMessage
          , ""
          , "<VirtualHost *:80>"
          , "    ServerAdmin " ++ serverAdmin
          , "    ServerName www." ++ domain
          , "    ServerAlias " ++ domain
          , ""
          , "    ErrorLog " ++ apacheErrorLog spec
          , "    CustomLog " ++ apacheAccessLog spec ++ " combined"
          , ""
          , "    ProxyRequests Off"
          , "    AllowEncodedSlashes NoDecode"
          , ""
          , "    <Proxy *>"
          , "                AddDefaultCharset off"
          , "                Order deny,allow"
          , "                #Allow from .example.com"
          , "                Deny from all"
          , "                #Allow from all"
          , "    </Proxy>"
          , ""
          , "    <Proxy http://127.0.0.1:" ++ port ++ "/*>"
          , "                AddDefaultCharset off"
          , "                Order deny,allow"
          , "                #Allow from .example.com"
          , "                #Deny from all"
          , "                Allow from all"
          , "    </Proxy>"
          , ""
          , "    ProxyPass / http://127.0.0.1:" ++ port ++ "/ nocanon"
          , "    ProxyPassReverse / http://127.0.0.1:" ++ port ++ "/"
          , "</VirtualHost>" ]
   else [])
  where
    port = fromMaybe (error $ "No port assignment for server type " ++ show package) (lookup package ports)

-- | A configuration file for the logrotate facility, installed via a line
-- in debianFiles.
logrotateConfig :: SiteSpec -> File
logrotateConfig spec@(SiteSpec {..}) =
    ("dist-ghc/logrotate" </> sitePackageName spec, apacheConfig ++ serverConfig)
    where
      apacheConfig =
          if needsApacheSite spec
          then [ apacheAccessLog spec ++ " {"
               , "  weekly"
               , "  rotate 5"
               , "  compress"
               , "  missingok"
               , "}"
               , apacheErrorLog spec ++ " {"
               , "  weekly"
               , "  rotate 5"
               , "  compress"
               , "  missingok"
               , "}" ]
          else []
      serverConfig =
          if needsInit spec
          then [ serverAccessLog spec ++ " {"
               , "  weekly"
               , "  rotate 5"
               , "  compress"
               , "  missingok"
               , "}"
               , serverAppLog spec ++ " {"
               , "  weekly"
               , "  rotate 5"
               , "  compress"
               , "  missingok"
               , "}" ]
          else []

-- | Given a SiteSpec template and a list of the package names from
-- the control file, generate a list of SiteSpec representing all the
-- debian configuration we need to generate.
testMulti :: SiteSpec -> [String] -> [SiteSpec]
testMulti spec packageNames =
    case nub (mapMaybe toPackage (filter (`elem` interestingPackageSuffixes) (mapMaybe (dropPrefix (siteName spec)) packageNames))) of
      [Server] -> [spec {package = Single}]
      xs | elem Server xs -> map (\ p -> spec {package = p}) xs
      _ -> error $ "No " ++ siteName spec ++ "-server package"
    where
      toPackage "-server" = Just Server
      toPackage "-production" = Just Production
      toPackage "-development" = Just Development
      toPackage "-staging" = Just Staging
      toPackage _ = Nothing

dropSuffix :: Eq a => [a] -> [a] -> Maybe [a]
dropSuffix suf list =
    if isSuffixOf suf list
    then Just (take (length list - length suf) list)
    else Nothing

dropPrefix :: Eq a => [a] -> [a] -> Maybe [a]
dropPrefix pre list =
    if isPrefixOf pre list
    then Just (drop (length pre) list)
    else Nothing

{-# DEPRECATED serverPackageName "use sitePackageName" #-}
serverPackageName :: SiteSpec -> String
serverPackageName = sitePackageName
